from .meta import *

import numpy as np


__all__ = [
  'get_number_of_parameters',
  'get_total_number_of_parameters',
  'graph_description'
]

def get_number_of_parameters(layer, **properties):
  params = get_parameters(layer, **properties)
  shapes = [ tuple(param().shape.as_list()) for param in params ]
  nums_weights = [ np.prod(shape, dtype='int64') for shape in shapes ]

  return sum(nums_weights)

def get_total_number_of_parameters(layer_or_layers, **properties):
  layers = get_layers(layer_or_layers)

  return sum([
    get_number_of_parameters(layer, **properties)
    for layer in layers
  ])

def get_attributes(layer, attributes):
  values = dict()

  for name, attr in attributes.items():
    if type(attr) is str:
      if hasattr(layer, attr):
        values[name] = getattr(layer, attr)
    elif callable(attr):
      values[name] = attr(layer)

  return values

def param_description(param):
  return '{name} {shape}: {properties}'.format(
    name=param.name if param.name is not None else param.__class__.__name__,
    shape=tuple(param().shape.as_list()),
    properties=', '.join(['%s=%s' % (k, v) for k, v in param.properties.items()])
  )

def layer_description(layer, output_shape, attributes, short=True):
  attrs = get_attributes(layer, attributes)
  additional = [ '%s=%s' % (k, v) for k, v in attrs.items() ]

  if short:
    return '{name}{clazz} ({nparams} params): {out_shape}{additional}'.format(
      name='' if layer.name is None else ('%s ' % layer.name),
      clazz=layer.__class__.__name__,
      out_shape=output_shape,
      nparams=get_number_of_parameters(layer),
      additional='' if len(additional) == 0 else ('\n  %s' % ', '.join(additional))
    )
  else:
    head = '{name}{clazz}: {out_shape}{additional}'.format(
      name='' if layer.name is None else ('%s ' % layer.name),
      clazz=layer.__class__.__name__,
      out_shape=output_shape,
      additional='' if len(additional) == 0 else ('\n  %s' % '\n  '.join(additional))
    )

    params = get_parameters(layer)
    if len(params) == 0:
      return head
    else:
      params_info ='\n  '.join([ param_description(param) for param in params ])

      total_number_of_params = 'number of params: %s' % get_number_of_parameters(layer)

      return '{head}\n  {params}\n  {total}'.format(
        head=head,
        params=params_info,
        total=total_number_of_params
      )

def graph_description(layer_or_layers, short=True, inputs=None, **attributes):
  all_layers = get_layers(layer_or_layers)

  if inputs is None:
    inputs = [
      layer for layer in all_layers
      if not hasattr(layer, 'incoming')
    ]

  outputs = layer_or_layers

  shapes = get_all_output_shapes(layer_or_layers)

  input_summary = ', '.join([
    ('%s' % shapes[layer]) if layer.name is None else ('%s: %s' % (layer.name, shapes[layer]))
    for layer in inputs
  ])

  output_summary = ', '.join([
    '{shape}'.format(shape=shapes[layer])
    for layer in outputs
  ])

  layer_delim = '\n' if short else '\n\n'

  layers_info = layer_delim.join([
    layer_description(layer, shapes[layer], attributes=attributes, short=short)
    for layer in all_layers
  ])

  summary = '{input_summary} -> {output_summary}\n{total}\n{sep}\n{layers_description}'.format(
    input_summary=input_summary,
    total='total number of params: %d' % get_total_number_of_parameters(layer_or_layers),
    sep='=' * 32,
    output_summary=output_summary,
    layers_description=layers_info
  )

  return summary


"""
General rules held in this module:
- params ending with `_op` like `conv_op` or simply op require function layer -> layer, e.g. conv(128, f=...)
- params not ending with `_op` like `conv` require function with similar signature as
    layer/subnetwork with the same name. Such parameters allow to override layer/subnetwork used.
    For example, overriding default nonlinearity: conv=lambda num_filters: conv(num_filters, f=T.nnet.sigmoid).
- usually, module provides two similar implementations of the same subnetwork like
  fire and fire_block. Subnetworks with shorter name are usually shortcuts to the method with longer name
  with popular default parameters.
"""

from .meta import *
from .selector import *

from .skip import *
from .residual_nets import *
from .vae_nets import *

from .normalization import *

from .meta_nets import *



import inspect

from ..layers import batch_norm
from ..nonlinearities import linear

__all__ = [
  'batch_normed'
]

def _batch_normed(layer, *bn_args, **bn_kwargs):
  signature = inspect.signature(layer)
  def new_layer(*args, **kwargs):
    parameters = signature.bind(*args, **kwargs)
    parameters.apply_defaults()
    arguments = parameters.arguments

    activation = arguments.pop('activation', linear())

    new_model = layer(**arguments, activation=linear())

    def f(*incoming):
      net = new_model(*incoming)
      return batch_norm(*bn_args, **bn_kwargs, activation=activation)(net)

    return f

  return new_layer

class BatchNormed(object):
  def __init__(self, *args, **kwargs):
    self.args = args
    self.kwargs = kwargs

  def __getitem__(self, layer):
    return _batch_normed(layer, *self.args, **self.kwargs)

batch_normed = BatchNormed


"""
CRAYNN is just another layer of abstraction for neural networks.

Important note: this package is poorly designed, unstable and lacks documentation.
"""

from .nonlinearities import *
from .regularization import *
from .parameters import *
from .updates import *
from .layers import *
from .subnetworks import *
from .networks import *

from . import train

from . import objectives

from . import utils

from . import viz

try:
  import crayflow as datasets
except:
  class DelayedWarning(object):
    def __getattr__(self, item):
      import warnings
      warnings.warn('`craynn.datasets` is now a separate package. Please, visit https://gitlab.com/craynn/crayflow')

      raise ModuleNotFoundError('datasets')

  datasets = DelayedWarning()



import tensorflow as tf

from ...parameters import default_weight_init, default_bias_init
from ...meta import derive
from ...nonlinearities import default_nonlinearity

from ..common import FunctionalLayer
from ..meta import get_output_shape, model_from

from .conv_utils import *

__all__ = [
  'Conv1DLayer', 'Conv2DLayer', 'Conv3DLayer',
  'PaddedConv1DLayer', 'PaddedConv2DLayer', 'PaddedConv3DLayer',
  'Conv1DLayer1x1', 'Conv2DLayer1x1', 'Conv3DLayer1x1',

  'conv_1d', 'conv_2d', 'conv_3d', 'conv',
  'pconv_1d', 'pconv_2d', 'pconv_3d', 'pconv',
  'conv_1d_1x1', 'conv_2d_1x1', 'conv_3d_1x1', 'conv_1x1'
]

class ConvLayer(FunctionalLayer):
  def __init__(self, incoming, ndim,
               num_filters,
               kernel_size=3,
               activation=default_nonlinearity,
               padding='valid', stride=1, dilation=1,
               W=default_weight_init,
               b=default_bias_init,
               name=None):
    super(ConvLayer, self).__init__(incoming, name=name)

    self.ndim = ndim

    input_shape = get_output_shape(incoming)

    if len(input_shape) != self.ndim + 2:
      raise ValueError(
        '%dD conv layer accepts only %dD tensors [got %s]!' % (self.ndim, self.ndim + 2, input_shape)
      )

    if num_filters <= 0:
      raise ValueError(
        '`num` channels must be > 0!'
      )

    self.input_channels = get_channel_dim(input_shape)
    self.input_size = get_spatial_dims(input_shape)
    self.num_channels = num_filters

    self.kernel_size = get_kernel_size(kernel_size, ndim=ndim)
    self.kernel_shape = self.kernel_size + (self.input_channels, self.num_channels)

    self.dilation = get_kernel_size(dilation, ndim=ndim)
    self.stride = get_kernel_size(stride, ndim=ndim)

    self.padding = padding.lower()
    self.parameters = [
      W(self.kernel_shape, name='W', weights=True, conv_kernel=True, trainable=True),
      b((self.num_channels,), name='b', biases=True, trainable=True)
    ]

    self.activation = activation

    self.conv_op = get_conv_op(ndim)

  def get_output_for(self, W, b, input, channel_last=False):
    if channel_last:
      input = to_channel_last(input, ndim=self.ndim)

    convolved = self.conv_op(
      input, W,
      strides=self.stride,
      padding=get_padding(self.padding),
      data_format=get_data_format(self.ndim, channel_last=channel_last),
      dilations=self.dilation,
      name=str(self) + '_conv%dd' % self.ndim,
    )

    if channel_last:
      convolved = to_channel_fist(convolved, self.ndim)

    broadcast = (None, ) + (slice(None), ) + (None, )* self.ndim
    return self.activation(
       convolved + b[broadcast]
    )

  def get_output_shape_for(self, input_shape):
    if len(input_shape) != self.ndim + 2:
      raise ValueError(
        '%dD conv layer accepts only %dD tensors [got %s]!' % (self.ndim, self.ndim + 2, input_shape)
      )

    return (input_shape[0], self.num_channels) + tuple([
      get_conv_output_shape(
        input_shape[i + 2],
        self.kernel_size[i],
        stride=self.stride[i],
        dilation=self.dilation[i],
        pad=self.padding
      )
      for i in range(self.ndim)
    ])


Conv1DLayer = derive('Conv1DLayer').based_on(ConvLayer).with_fixed(ndim=1)
Conv2DLayer = derive('Conv2DLayer').based_on(ConvLayer).with_fixed(ndim=2)
Conv3DLayer = derive('Conv3DLayer').based_on(ConvLayer).with_fixed(ndim=3)

conv_1d = model_from(Conv1DLayer)()
conv_2d = model_from(Conv2DLayer)()
conv_3d = model_from(Conv3DLayer)()

conv = select_by_dim([None, conv_1d, conv_2d, conv_3d])

Conv1DLayer1x1 = derive('Conv1DLayer1x1').based_on(ConvLayer).with_fixed(ndim=1, kernel_size=1)
Conv2DLayer1x1 = derive('Conv2DLayer1x1').based_on(ConvLayer).with_fixed(ndim=2, kernel_size=1)
Conv3DLayer1x1 = derive('Conv3DLayer1x1').based_on(ConvLayer).with_fixed(ndim=3, kernel_size=1)

conv_1d_1x1 = model_from(Conv1DLayer1x1)()
conv_2d_1x1 = model_from(Conv2DLayer1x1)()
conv_3d_1x1 = model_from(Conv3DLayer1x1)()

conv_1x1 = select_by_dim([None, conv_1d_1x1, conv_2d_1x1, conv_3d_1x1])

conv_1d = model_from(Conv1DLayer)()
conv_2d = model_from(Conv2DLayer)()
conv_3d = model_from(Conv3DLayer)()


PaddedConv1DLayer = derive('PaddedConv1DLayer').based_on(Conv1DLayer).with_fixed(padding='same')
PaddedConv2DLayer = derive('PaddedConv2DLayer').based_on(Conv2DLayer).with_fixed(padding='same')
PaddedConv3DLayer = derive('PaddedConv3DLayer').based_on(Conv3DLayer).with_fixed(padding='same')

pconv_1d = model_from(PaddedConv1DLayer)()
pconv_2d = model_from(PaddedConv2DLayer)()
pconv_3d = model_from(PaddedConv3DLayer)()

pconv = select_by_dim([None, pconv_1d, pconv_2d, pconv_3d])


import numpy as np
import tensorflow as tf

from ..meta import derive
from .meta import FunctionalLayer, model_from, get_output_shape

__all__ = [
  'NoiseLayer', 'noise',
  'GaussianNoiseLayer', 'gaussian_noise',

  'GaussianRandomVariableLayer', 'gaussian_rv',
  'DropoutLayer', 'dropout'
]

class NoiseLayer(FunctionalLayer):
  def __init__(self, incoming, noise_op, std=1.0e-3, name=None):
    super(NoiseLayer, self).__init__(incoming, name=name)

    self.noise_op = noise_op
    self.std = std

  def get_output_for(self, input, deterministic=False):
    if deterministic:
      return input
    else:
      return input + self.noise_op(shape=tf.shape(input), dtype=input.dtype) * self.std

  def get_output_shape_for(self, input_shape):
    return input_shape


noise = model_from(NoiseLayer)()

GaussianNoiseLayer = derive('GaussianNoiseLayer').based_on(NoiseLayer).with_fixed(noise_op=tf.random.normal)
gaussian_noise = model_from(GaussianNoiseLayer)()

class DropoutLayer(FunctionalLayer):
  def __init__(self, incoming, p=0.2, name=None):
    super(DropoutLayer, self).__init__(incoming, name=name)
    self.p = p

  def get_output_for(self, input, deterministic=False):
    if deterministic:
      return input
    else:
      return tf.nn.dropout(input, rate=self.p, noise_shape=None, name=self.name)

  def get_output_shape_for(self, input_shape):
    return input_shape

dropout = model_from(DropoutLayer)()

class GaussianRandomVariableLayer(FunctionalLayer):
  def __init__(self, *incoming, name=None):
    assert len(incoming) == 2

    super(GaussianRandomVariableLayer, self).__init__(*incoming, name=name)

  def get_output_for(self, *inputs, deterministic=False):
    if deterministic:
      return inputs[0]
    else:
      mean, std = inputs
      u = tf.random.normal(shape=mean.shape, dtype=mean.dtype)
      return u * std + mean

  def get_output_shape_for(self, *input_shapes):
    return input_shapes[0]

gaussian_rv = model_from(GaussianRandomVariableLayer)()




import tensorflow as tf

__all__ = [
  'combine_properties',
  'init_name'
]

def combine_properties(manual_properties, default_properties):
  props = default_properties.copy()
  props.update(manual_properties)

  return props

init_name = lambda name: name + '_init' if name is not None else None

def copy_variable(variable : tf.Variable, validate_shape=False, **kwargs):
  return tf.Variable(
    initial_value=variable.initial_value,
    dtype=variable.dtype,
    trainable=variable.trainable,
    validate_shape=validate_shape,
    **kwargs
  )


import tensorflow as tf

from ..parameters import default_weight_init, default_bias_init

from .meta import FunctionalLayer, get_output_shape, model_from
from ..nonlinearities import default_semibounded_nonlinearity


__all__ = [
  'DenseLayer',
  'dense',

  'TensorDenseLayer',
  'tensor_dense',

  'BatchDenseLayer',
  'batch_dense',
]

class DenseLayer(FunctionalLayer):
  def __init__(self, incoming, num_units,
               activation=default_semibounded_nonlinearity,
               W=default_weight_init,
               b=default_bias_init,
               name=None):
    super(DenseLayer, self).__init__(incoming, name=name)
    input_shape = get_output_shape(incoming)
    self.num_units = num_units

    if len(input_shape) != 2:
      raise ValueError('Dense layer accepts only 2D tensors got [%s]!' % input_shape)

    self.parameters = [
      W(shape=(input_shape[1], num_units), name='W', weights=True, trainable=True),
      b(shape=(num_units,), name='b', biases=True, trainable=True)
    ]

    self.activation = activation

  def get_output_for(self, W, b, input):
    return self.activation(
      tf.matmul(input, W) + b[None, :]
    )

  def get_output_shape_for(self, input_shape):
    if len(input_shape) != 2:
      raise ValueError('Dense layer accepts only 2D tensors!')

    return (input_shape[0], self.num_units)


dense = model_from(DenseLayer).with_fixed().with_defaults()()

class TensorDenseLayer(FunctionalLayer):
  def __init__(self, incoming, num_units,
               activation=default_semibounded_nonlinearity,
               W=default_weight_init,
               b=default_bias_init,
               axis=-1,
               name=None):
    super(TensorDenseLayer, self).__init__(incoming, name=name)
    input_shape = get_output_shape(incoming)
    self.num_units = num_units

    self.axis = (len(input_shape) + axis) % len(input_shape)

    self.parameters = [
      W(shape=(input_shape[self.axis], num_units), name='W', weights=True, trainable=True),
      b(shape=(num_units,), name='b', biases=True, trainable=True)
    ]

    self.b_broadcast = [ (None if i != self.axis else slice(None, None, None)) for i in range(len(input_shape)) ]

    self.activation = activation

  def get_output_for(self, W, b, input):
    return self.activation(
      tf.tensordot(input, W, axes=[(self.axis, ), (0, )]) + b[self.b_broadcast]
    )

  def get_output_shape_for(self, input_shape):
    return tuple([
      (input_shape[i] if i != self.axis else self.num_units)
      for i in range(len(input_shape))
    ])


tensor_dense = model_from(TensorDenseLayer).with_fixed().with_defaults()()


class BatchDenseLayer(FunctionalLayer):
  def __init__(self, incoming, num_units, num_batches,
               activation=default_semibounded_nonlinearity,
               W=default_weight_init,
               b=default_bias_init,
               axis=-1,
               name=None):
    super(BatchDenseLayer, self).__init__(incoming, name=name)
    input_shape = get_output_shape(incoming)
    self.num_batches = num_batches
    self.num_units = num_units

    self.axis = (len(input_shape) + axis) % len(input_shape)

    self.parameters = [
      W(shape=(input_shape[self.axis], num_batches, num_units), name='W', weights=True, trainable=True),
      b(shape=(num_batches, num_units,), name='b', biases=True, trainable=True)
    ]

    self.b_broadcast = tuple([None] * (len(input_shape) - 1) + [slice(None, None, None)] * 2)

    self.activation = activation

  def get_output_for(self, W, b, input):
    return self.activation(
      tf.tensordot(input, W, axes=[(self.axis, ), (0, )]) + b[self.b_broadcast]
    )

  def get_output_shape_for(self, input_shape):
    return tuple([
      input_shape[i]
      for i in range(len(input_shape))
      if i != self.axis
    ]) + (self.num_batches, self.num_units)


batch_dense = model_from(BatchDenseLayer).with_fixed().with_defaults()()


import tensorflow as tf

__all__ = [
  'GradientOptimizer'
]

class GradientOptimizer(object):
  def __init__(self, target, variables):
    self.target = target
    self.variables = variables

    self.tape = None

  def __enter__(self):
    if self.tape is not None:
      raise RuntimeError('Reentering optimizer context')

    self.tape = tf.GradientTape(persistent=False, watch_accessed_variables=False)
    self.tape.__enter__()
    self.tape.watch(self.variables)

  def __exit__(self, exc_type, exc_val, exc_tb):
    self.tape.__exit__(exc_type, exc_val, exc_tb)
    self.tape = None

  def apply_gradients(self, gradients):
    raise NotImplementedError()

  def step(self, *args):
    if self.tape is None:
      raise RuntimeError('step is called without entering optimizer context, please, consider using __call__ instead')

    value = self.target(*args)
    if isinstance(value, tuple) or isinstance(value, list):
      value_ = sum(value)
    else:
      value_ = value

    gradients = self.tape.gradient(value_, self.variables)
    self.apply_gradients(gradients)

    return value

  @tf.function
  def __call__(self, *args):
    with self:
      return self.step(*args)


import tensorflow as tf

from .common import zeros_init, glorot_normal_init
from .meta import Parameter, parameter_model

__all__ = [
  'SoftmaxParameter',
  'softmax_parameter',

  'PositiveParameter',
  'positive_parameter',
]

class SoftmaxParameter(Parameter):
  def __init__(self, shape, scale=1, w=zeros_init(), properties=None, name=None):
    if properties is None:
      properties = dict()

    self.w = w(shape=shape, **properties, name=(name + '_w') if name is not None else None)
    self.activation = tf.nn.softmax(self.w(), axis=0)

    if scale != 1:
      self.activation = self.activation * scale
    
    super(SoftmaxParameter, self).__init__(shape, properties, name=name)

  def __call__(self):
    return self.activation

  def dependencies(self):
    return [self.w]

  def own_variables(self):
    return []

softmax_parameter = parameter_model(SoftmaxParameter)


class PositiveParameter(Parameter):
  def __init__(self, shape, w=glorot_normal_init(gain=0.5), properties=None, name=None):
    if properties is None:
      properties = dict()

    self.w = w(shape=shape, **properties, name=(name + '_w') if name is not None else None)
    self.activation = tf.nn.softplus(self.w())

    super(PositiveParameter, self).__init__(shape, properties, name=name)

  def __call__(self):
    return self.activation

  def dependencies(self):
    return [self.w]

  def own_variables(self):
    return []


positive_parameter = parameter_model(PositiveParameter)


### inspired by
### https://github.com/Lasagne/Lasagne/blob/master/lasagne/layers/normalization.py#L124-L326

import tensorflow as tf

from ..nonlinearities import linear
from .meta import FunctionalLayer, get_output_shape, model_from
from ..parameters import zeros_init, ones_init, const_init

__all__ = [
  'BatchNormLayer',
  'batch_norm',

  'LayerNormLayer',
  'layer_norm',

  'RollingNormalizationLayer',
  'rolling_normalization',

  'NormalizationLayer',
  'norm'
]

class BatchNormLayer(FunctionalLayer):
  def __init__(self, incoming, gamma=ones_init(), beta=zeros_init(),
               axes=(0, ), activation=linear(), epsilon=1.0e-4, name=None):
    self.epsilon = tf.constant(epsilon, dtype='float32')

    incoming_shape = get_output_shape(incoming)
    if axes == 'auto':
      axes = list(range(len(incoming_shape) - 1))
    self.axes = axes

    self.parameter_shape = tuple([
      d
      for i, d in enumerate(incoming_shape)
      if i not in self.axes
    ])

    self.broadcast = tuple([
      (None if i in axes else slice(None, None, None)) for i, _ in enumerate(incoming_shape)
    ])

    self.parameters = []

    if gamma is not None:
      self.parameters.append(
        gamma(self.parameter_shape, weights=True, normalization_scales=True, trainable=True)
      )

    if beta is not None:
      self.parameters.append(
        beta(self.parameter_shape, biases=True, normalization_biases=True, trainable=True)
      )

    self.scale = gamma is not None
    self.bias = beta is not None

    self.activation = activation

    super(BatchNormLayer, self).__init__(incoming, name=name)

  def get_output_for(self, *args):
    input = args[-1]

    if self.scale:
      gamma = args[0]
    else:
      gamma = None

    if self.bias:
      beta = args[1] if self.scale else args[0]
    else:
      beta = None

    input_mean = tf.reduce_mean(input, axis=self.axes)
    input_var = tf.reduce_mean((input - input_mean[self.broadcast]) ** 2, axis=self.axes)
    input_inverse_std = 1 / tf.sqrt(input_var + self.epsilon)

    y = (input - input_mean[self.broadcast]) * input_inverse_std[self.broadcast]

    scaled = y if gamma is None else y * gamma[self.broadcast]
    biased = scaled if beta is None else scaled + beta[self.broadcast]

    return self.activation(biased)

  def get_output_shape_for(self, input_shape):
    return input_shape

batch_norm = model_from(BatchNormLayer)()


class LayerNormLayer(FunctionalLayer):
  def __init__(self, incoming, axes='auto', epsilon=1.0e-4, name=None):
    self.epsilon = tf.constant(epsilon, dtype='float32')

    incoming_shape = get_output_shape(incoming)
    if axes == 'auto':
      axes = tuple(range(1, len(incoming_shape) - 1))

    self.axes = axes

    self.broadcast = tuple([
      (tf.newaxis if i in axes else slice(None, None, None)) for i, _ in enumerate(incoming_shape)
    ])

    super(LayerNormLayer, self).__init__(incoming, name=name)

  def get_output_for(self, input):
    input_mean = tf.reduce_mean(input, axis=self.axes)
    input_var = tf.reduce_mean((input - input_mean[self.broadcast]) ** 2, axis=self.axes)
    input_inverse_std = 1 / tf.sqrt(input_var + self.epsilon)

    return (input - input_mean[self.broadcast]) * input_inverse_std[self.broadcast]

  def get_output_shape_for(self, input_shape):
    return input_shape


layer_norm = model_from(LayerNormLayer)()

class RollingNormalizationLayer(FunctionalLayer):
  def __init__(self, incoming, gamma=ones_init(), beta=zeros_init(),
               axes=(0,), rho=0.99, epsilon=1e-2, name=None):
    self.epsilon = tf.constant(epsilon, dtype='float32')
    self.rho = tf.constant(rho, dtype='float32')
    self.crho = tf.constant(1 - rho, dtype='float32')

    incoming_shape = get_output_shape(incoming)
    if axes == 'auto':
      axes = list(range(len(incoming_shape) - 1))
    self.axes = axes

    self.parameter_shape = tuple([
      d
      for i, d in enumerate(incoming_shape)
      if i not in self.axes
    ])

    self.broadcast = tuple([
      (None if i in axes else slice(None, None, None)) for i, _ in enumerate(incoming_shape)
    ])

    self.parameters = [
      gamma(self.parameter_shape, weights=True, normalization_scales=True, trainable=False),
      beta(self.parameter_shape, biases=True, normalization_biases=True, trainable=False)
    ]

    super(RollingNormalizationLayer, self).__init__(incoming, name=name)

  def get_output_for(self, gamma, beta, input, deterministic=False, normalization=True):
    if not deterministic and normalization:
      input_mean = tf.reduce_mean(input, axis=self.axes)
      beta.assign(
        self.rho * beta + self.crho * input_mean
      )

      input_var = tf.reduce_mean((input - input_mean[self.broadcast]) ** 2, axis=self.axes)
      input_inverse_std = 1 / tf.sqrt(input_var + self.epsilon)

      gamma.assign(
        self.rho * gamma + self.crho * input_inverse_std
      )

    return input * gamma[self.broadcast] + beta[self.broadcast]

  def get_output_shape_for(self, input_shape):
    return input_shape

rolling_normalization = model_from(RollingNormalizationLayer)()


class NormalizationLayer(FunctionalLayer):
  def __init__(self, incoming, gamma=const_init(value=0.5), beta=zeros_init(),
               axes=(0,), name=None):

    incoming_shape = get_output_shape(incoming)
    if axes == 'auto':
      axes = list(range(len(incoming_shape) - 1))
    self.axes = axes

    self.parameter_shape = tuple([
      d
      for i, d in enumerate(incoming_shape)
      if i not in self.axes
    ])

    self.broadcast = tuple([
      (None if i in axes else slice(None, None, None)) for i, _ in enumerate(incoming_shape)
    ])

    self.parameters = [
      gamma(self.parameter_shape, weights=True, normalization_scales=True, trainable=True),
      beta(self.parameter_shape, biases=True, normalization_biases=True, trainable=True)
    ]

    super(NormalizationLayer, self).__init__(incoming, name=name)

  def get_output_for(self, gamma, beta, input):
    return input * tf.nn.softplus(gamma)[self.broadcast] + beta[self.broadcast]

  def get_output_shape_for(self, input_shape):
    return input_shape

norm = model_from(NormalizationLayer)()


"""
CRAYNN - yet another NN toolkit.
"""

from setuptools import setup, find_packages
import os

here = os.path.dirname(__file__)

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
  long_description=f.read()

with open(os.path.join(here, 'VERSION'), encoding='utf-8') as f:
  version=f.read().strip()

setup(
  name = 'craynn',
  version=version,
  description="""Yet another neural network toolkit.""",

  long_description=long_description,
  long_description_content_type="text/markdown",

  url='https://github.com/craynn/craynn',

  author='Maxim Borisyak and contributors.',
  author_email='maximus.been@gmail.com',

  maintainer='Maxim Borisyak',
  maintainer_email='maximus.been@gmail.com',

  license='MIT',

  classifiers=[
    'Development Status :: 4 - Beta',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
  ],

  packages=find_packages(exclude=['contrib', 'examples', 'docs', 'tests']),

  extras_require={
    'test': [
      'pytest>=4.0.0',
      'scipy>=1.3.0'
    ],
  },

  install_requires=[
    'tensorflow>=2.0.0',
    'numpy>=1.17.1',
    'matplotlib>=3.0.2',
    'pydotplus>=2.0.2',
  ],

  python_requires='>=3.7',
)





from inspect import signature

__all__ = [
  'get_kwargs',
  'apply_with_kwargs'
]

def get_kwargs(func):
  params = signature(func).parameters
  return [p.name for p in params.values()]

def apply_with_kwargs(f, *args, **kwargs):
  accepted_kwargs = get_kwargs(f)
  passed_kwargs = dict()

  for k, v in kwargs.items():
    if k in accepted_kwargs:
      passed_kwargs[k] = v

  return f(*args, **passed_kwargs)


from ..nonlinearities import default_nonlinearity
from ..parameters import unbound_parameter
from ..layers import MetaLayer

from . import achain

__all__ = [
  'meta_dense',
  'meta_conv'
]

def meta_dense(num_units, meta_ops, activation=default_nonlinearity, name=None):
  from ..parameters import glorot_normal_init
  from ..layers import dense, glorot_dense, reshape, transpose

  def f(incoming):
    op = dense(
      num_units, activation=activation,
      W=unbound_parameter(),
      b=unbound_parameter()
    )(incoming)

    W, b = op.parameters
    W_shape, b_shape = W.shape, b.shape

    W_layer, b_layer = achain(
      meta_ops, [
        (glorot_dense(W_shape[0], target_shape=W_shape), transpose([1, 0])),
        (dense(1, W=glorot_normal_init(gain=0.1)), reshape(b_shape)),
      ]
    )()

    return MetaLayer((incoming, ), (W_layer, b_layer), op=op, name=name)

  return f

def meta_conv(
  num_filters, kernel_size=3, activation=default_nonlinearity,
  pad='valid', stride=1, dilation=1,
  meta_ops=None, name=None
):

  from ..parameters import glorot_normal_init
  from ..layers import conv, reshape, transpose, dense, glorot_dense

  def f(incoming):
    op = conv(
      num_filters, kernel_size=kernel_size,
      pad=pad, stride=stride, dilation=dilation,
      activation=activation,
      W=unbound_parameter(),
      b=unbound_parameter()
    )(incoming)

    W, b = op.parameters
    W_shape, b_shape = W.shape, b.shape
    W_shape_flat = W_shape[0] * W_shape[1] * W_shape[2]

    W_layer, b_layer = achain(
      meta_ops, [
        (glorot_dense(W_shape_flat, target_shape=W_shape), transpose([1, 0]), reshape(W_shape)),
        (dense(1, W=glorot_normal_init(gain=0.1)), reshape(b_shape)),
      ]
    )()

    return MetaLayer((incoming, ), (W_layer, b_layer), op=op, name=name)

  return f



from .meta import FunctionalLayer

__all__ = [
  'MetaLayer', 'meta',

  'binary_const_input',
  'binary_var_input',
  'glorot_dense'
]

class MetaLayer(FunctionalLayer):
  def __init__(self, incoming, incoming_parameters, op, name=None):
    """
    Redirects outputs of `incoming_parameters` to parameters of `incoming` layer.
    :param incoming: tuple of incoming data layers
    :param incoming_parameters: tuple of incoming parameter layers
    :param op: operation to wrap
    """
    if name is None:
      name = 'Meta[%s]' % (op.__class__.__name__ if op.name is None else op.name, )

    super(MetaLayer, self).__init__(*incoming, *incoming_parameters, name=name)

    self.incoming_data = incoming
    self.incoming_parameters = incoming_parameters

    self.op = op
    self.parameters = []

  def get_output_shape_for(self, *input_shapes, **modes):
    data_shapes = input_shapes[:len(self.incoming_data)]
    return self.op.get_output_shape_for(*data_shapes, **modes)

  def get_output_for(self, *inputs, **modes):
    n = len(self.incoming_data)
    data = inputs[:n]
    params = inputs[n:]
    return self.op.get_output_for(*params, *data, **modes)

class Meta(object):
  def __call__(self, op):
    return lambda incoming_input, *incoming_params, name=None: \
      MetaLayer(incoming_input, incoming_params, op=op, name=name)

  def __getitem__(self, op):
    return self(op)

meta = Meta()

from .common import const_input, variable_input
from ..utils import binary_encoding
from ..parameters import const_init

binary_const_input = lambda n, name=None: const_input(binary_encoding(range(n), size=n), name=name)

def binary_var_input(n, name=None):
  value = binary_encoding(range(n), size=n)
  return variable_input(shape=value.shape, V=const_init(value=value), name=name)

from ..layers import dense
from ..nonlinearities import linear
from ..parameters import glorot_normal_double_init, zeros_init


glorot_dense = lambda num_units, target_shape, name=None: \
  dense(
    num_units,
    W=glorot_normal_double_init(gain=0.1, target_shape=target_shape),
    b=zeros_init(),
    activation=linear(),
    name=name
  )


__all__ = [
  'many_to_one',
  'one_to_one',
  'others',
  'default_mapping',

  'encode_text', 'encode_text_groups'
]

class TextMapping(object):
  def table(self):
    import numpy as np
    return np.array([
      self.translate(chr(i))
      for i in range(256)
    ], dtype='uint8')

  def __len__(self):
    raise NotImplementedError()

  def __or__(self, other):
    return UnionMapping(self, other)

  def __add__(self, other):
    return ConcatMapping(self, other)

  def translate(self, character):
    raise NotImplementedError()

  def __call__(self, text):
    return [
      self.translate(c)
      for c in text
    ]

class ConcatMapping(TextMapping):
  def __init__(self, first : TextMapping, second : TextMapping):
    self._first = first
    self._second = second
    self._offset = len(self._second)

  def __len__(self):
    return len(self._first) + len(self._second)

  def translate(self, character):
    result = self._first.translate(character)
    if result is not None:
      return result + self._offset
    else:
      result = self._second.translate(character)
      return result

class UnionMapping(TextMapping):
  def __init__(self, first : TextMapping, second : TextMapping):
    self._first = first
    self._second = second

  def __len__(self):
    return max(len(self._first), len(self._second))

  def translate(self, character):
    result = self._first.translate(character)
    if result is None:
      result = self._second.translate(character)

    return result

class ManyToOneMapping(TextMapping):
  def __init__(self, characters):
    self._characters = set(characters)
    super(ManyToOneMapping, self).__init__()

  def __len__(self):
    return 1

  def translate(self, character):
    return 0 if character in self._characters else None

class OneToOneMapping(TextMapping):
  def __init__(self, characters):
    self._mapping = dict(zip(characters, range(len(characters))))

  def __len__(self):
    return len(self._mapping)

  def translate(self, character):
    return self._mapping.get(character, None)

class EtcMapping(TextMapping):
  def __init__(self):
    super(EtcMapping, self).__init__()

  def __len__(self):
    return 1

  def translate(self, character):
    return 0

many_to_one = ManyToOneMapping
one_to_one = OneToOneMapping
others = EtcMapping

import string
default_mapping = many_to_one(string.whitespace) + (
  one_to_one(string.ascii_lowercase) | one_to_one(string.ascii_uppercase)
) + others()

def encode_text(mapping):
  import numpy as np
  try:
    table = mapping.table()
  except:
    table = mapping

  def f(s):
    array = np.array([ ord(c) for c in s ], dtype='uint8')
    return table[array]

  return f

def encode_text_groups(mapping, fill=' '):
  import numpy as np

  try:
    table = mapping.table()
  except:
    table = mapping

  def f(group):
    group_len = max([ len(s) for s in group ])
    encoded = np.ones(shape=(len(group), group_len), dtype='uint8') * table[ord(fill)]

    for i, s in enumerate(group):
      array = np.array([ord(c) for c in s], dtype='int64')
      encoded[i, :len(s)] = table[array]

    return encoded

  return f


from .meta import achain
from ..layers import *

def residual_connection(*body, merge_op=elementwise_mean()):
  def constructor(incoming):
    origin = incoming
    net = achain(*body)(incoming)
    return merge_op(origin, net)

  return constructor
